from __future__ import annotations
import logging, csv
from pathlib import Path
from django.utils import timezone
from ..format import slugen
from . import get_connection, call_procedure, truncate_table

logger = logging.getLogger(__name__)

def _compute_pathes(sql_path: Path, csv_path: Path|str = None) -> Path:
    if not isinstance(sql_path, Path):
        sql_path = Path(sql_path)

    if not csv_path:
        csv_path = sql_path.parent.joinpath(sql_path.stem + ".local.csv")
    else:
        if not isinstance(csv_path, Path):
            csv_path = Path(csv_path)
        if csv_path.is_dir():
            csv_path = csv_path.joinpath(sql_path.stem + ".local.csv")

    csv_path.parent.mkdir(parents=True, exist_ok=True)

    return sql_path, csv_path


def extract_from_sql(sql_path: Path|str, connection: str|object = None,
    csv_path: Path|str = None,
    delimiter = ";", encoding="utf-8") -> int:
    """
    Extract data using a query file to a CSV file.
    - `sql_path`: file containing the extract query.
    - `connection`: source connection. If `None`, use default Django connection.
    """
    sql_path, csv_path = _compute_pathes(sql_path, csv_path)
    connection = get_connection(connection=connection)

    logger.info(f"extract {sql_path} to {csv_path}")

    # For CSV file:
    # - Set newline to '', otherwise newlines embedded inside quoted fields will not be interpreted correctly. See footnote of: https://docs.python.org/3/library/csv.html
    # - Set encoding to utf-8-sig (UTF8 with BOM): CSV is for exchanges, encoding should not depend on the exporting operating system. BOM is necessary for correct display with Excel
    with open(csv_path, "w", newline="", encoding="utf-8-sig" if encoding == "utf-8" else encoding) as file:
        writer = csv.writer(file, delimiter=delimiter)
        with connection.cursor() as cursor:
            cursor.execute(sql_path.read_text())

            writer.writerow([column[0] for column in cursor.description])
            row = cursor.fetchone()
            while row: 
                writer.writerow([value for value in row])
                row = cursor.fetchone()


def load_from_csv(path: Path|str, model: type,
    truncate: bool = True, integrate: str|list[str] = None,
    delimiter = ";", encoding = "utf-8", accept_ignored_headers = True, mapping = None, static_mapping = None) -> int:
    """
    Load from CSV file `path` to `model` class.
    
    Model class must be defined with CopyManager. Example:

    ```py
    from django.db import models
    from postgres_copy import CopyManager

    class MyModel(models.Model):
        objects = CopyManager()
    ```
    """
    if isinstance(path, str):
        path = Path(path)
        
    if not hasattr(model, "objects") or not hasattr(model.objects, "from_csv"):    
        raise ValueError("missing %s.objects.from_csv, did you specified `objects = CopyManager()`?" % model.__qualname__)

    db_table = model._meta.db_table
    logger.info("load %s in table %s (model %s)", path, db_table, model.__name__)

    fields = [field.name for field in model._meta.get_fields()]

    # Change encoding to utf-8-sig if file starts with UTF8-BOM
    if encoding == "utf-8":
        with open(path, mode="r", encoding="utf-8") as file:
            data = file.read(1)
            if data == "\ufeff":
                encoding = "utf-8-sig"

    # Get CSV headers
    headers = []
    with open(path, newline="", encoding=encoding) as file:
        reader = csv.reader(file, delimiter=delimiter)
        for row in reader:
            headers = row
            break

    if not headers:
        raise ValueError("headers not found in %s", path)

    # Build mapping
    if mapping is None:
        mapping = {}

    def search_field(name, lowersearch):
        """ Returns True to continue headers loop"""
        for field in fields:
            if field.lower() == lowersearch:
                if field in mapping:
                    logger.warning("ignore header \"%s\": cannot map to field \"%s\" (already added mapped to header \"%s\")", name, field, mapping[field])                    
                    return True
                mapping[field] = name
                return True
        return False

    ignored_headers = ""
    for name in headers:
        # Try to find field using lowercase
        if search_field(name, name.lower()):
            continue

        # Try to find field using slug
        slug = slugen(name, separator="_")
        if search_field(name, slug):
            continue

        # Not found in fields
        ignored_headers += (", " if ignored_headers else "") + name + (f" ({slug})" if slug != name else "")

    if ignored_headers:
        logger.log(logging.INFO if accept_ignored_headers else logging.WARNING, "headers ignored in %s: %s", path.name, ignored_headers)

    if static_mapping is None:
        static_mapping = {}
    if "load_at" in fields and not "load_at" in static_mapping:
        static_mapping["load_at"] = timezone.now()
   
    # Truncate table
    if truncate:
        truncate_table(model)

    # Load
    with open(path, newline="", encoding=encoding) as file:
        insert_count = model.objects.from_csv(file, mapping=mapping, static_mapping=static_mapping, delimiter=delimiter)
        logger.info("%d records loaded in %s", insert_count, db_table)

    # Integrate
    if integrate:
        if not isinstance(integrate, list):
            integrate = [integrate]

        dst_connection = get_connection(model=model)
        call_procedure(names=integrate, connection=dst_connection)

    return insert_count


def transfer_from_sql(sql_path: Path|str, model: type, connection: str|object = None,
    truncate: bool = True, integrate: str|list[str] = None,
    csv_path: Path|str = None,
    skip_extract = False, skip_load = False,
    delimiter = ";", encoding="utf-8", accept_ignored_headers = True, mapping = None, static_mapping = None) -> int|None:
    """
    Extract data using a query file to a CSV file, and load the CSV file to a model class.
    - `sql_path`: file containing the extract query.
    - `model`: target Model.
    - `connection`: source connection (used during extract from `sql_path`). If `None`, use default Django connection.
    """
    sql_path, csv_path = _compute_pathes(sql_path, csv_path)

    result = None

    if not skip_extract:
        extract_from_sql(sql_path, connection=connection, csv_path=csv_path, delimiter = ";", encoding="utf-8")

    if not skip_load and model:
        result = load_from_csv(csv_path, model, truncate=truncate, integrate=integrate, delimiter=delimiter, encoding=encoding, accept_ignored_headers=accept_ignored_headers, mapping=mapping, static_mapping=static_mapping)
        
    return result



from __future__ import annotations
import os, sys
from datetime import datetime
from getpass import getpass

if sys.platform == "win32":
    from keyring.backends.Windows import WinVaultKeyring
    import win32cred

    # See: https://docs.microsoft.com/en-us/windows/win32/api/wincred/ns-wincred-credentiala
    CRED_TYPE_GENERIC = 1 # The credential is a generic credential. The credential will not be used by any particular authentication package. The credential will be stored securely but has no other significant characteristics.
    CRED_TYPE_DOMAIN_PASSWORD = 2 # The credential is a password credential and is specific to Microsoft's authentication packages. The NTLM, Kerberos, and Negotiate authentication packages will automatically use this credential when connecting to the named target.
    CRED_TYPE_DOMAIN_CERTIFICATE = 3 # The credential is a certificate credential and is specific to Microsoft's authentication packages. The Kerberos, Negotiate, and Schannel authentication packages automatically use this credential when connecting to the named target.

    enumerate_additional_props = ["LastWritten", "TargetAlias", "Comment", "Type", "Flags", "CredentialBlob"]
    enumerate_headers = ["Service", "Username", *enumerate_additional_props]

    def enumerate(service=None, username=None):
        if service is not None:
            service = service.lower()
        if username is not None:
            username = username.lower()
        
        for cred in win32cred.CredEnumerate():
            cred_service = cred["TargetName"]
            cred_username = cred["UserName"]

            # Filter
            if service is not None:
                if not service in cred_service.lower():
                    continue
            if username is not None:
                if cred_username is None or not username in cred_username.lower():
                    continue

            # Additional properties
            props = []
            for prop in enumerate_additional_props:
                if prop == "CredentialBlob":
                    props.append("X" if cred[prop] else "")
                elif prop == "Type":
                    indication = ""
                    if cred["Type"] == CRED_TYPE_GENERIC:
                        indication = "generic"
                    elif cred["Type"] == CRED_TYPE_DOMAIN_PASSWORD:
                        indication = "domain password"
                    elif cred["Type"] == CRED_TYPE_DOMAIN_CERTIFICATE:
                        indication = "domain certificate"
                    props.append(str(cred[prop]) + (" (%s)" % indication if indication else ""))
                else:
                    props.append(cred[prop])

            yield [cred_service, cred_username, *props]
    
    def get_username(service):
        for cred in win32cred.CredEnumerate():
            if cred["Type"] == CRED_TYPE_GENERIC and cred["TargetName"] == service and cred["UserName"]:
                return cred["UserName"]

        # not found in credential manager
        return None


    _keyring = None

    def _get_keyring():
        global _keyring
        if _keyring is None:
            _keyring = WinVaultKeyring()
        return _keyring

    def get_password(service, username):
        return _get_keyring().get_password(service, username)


    def set_password(service, username, password):
        return _get_keyring().set_password(service, username, password)


    def delete_password(service, username):
        return _get_keyring().delete_password(service, username)


else:
    import re, subprocess
    from glob import glob
    from ..format import SubprocessError

    _several_slashes_re = re.compile(r"/{2,}")

    def _sanitize_part(part):
        """ Replace several slashes with only one """
        if not part:
            return part
        part = part.replace("://", ":")
        return _several_slashes_re.sub("/", part)
        
    def _get_key(service, username):
        return os.path.join(
            _sanitize_part(service),
            _sanitize_part(username),
        ).rstrip("/")
    
    enumerate_headers = ["Service", "Username", "LastWritten", "Path"]

    def enumerate(service=None, username=None):
        if service is not None:
            service = _sanitize_part(service.lower())
        if username is not None:
            username = _sanitize_part(username.lower())

        prefix = os.path.expanduser("~/.password-store/")
        for path in glob(f"{prefix}**/*.gpg", recursive=True):
            name_without_extension = path[:-4]
            name_parts = name_without_extension[len(prefix):].rsplit("/", 1)
            if len(name_parts) == 0:             
                cred_service = ""
                cred_username = ""
            elif len(name_parts) == 1:                
                cred_service = name_parts[0]
                cred_username = ""
            else:
                cred_service = "/".join(name_parts[:-1])
                cred_username = name_parts[-1]

            # Filter
            if service is not None:
                if not service in cred_service.lower():
                    continue
            if username is not None:
                if cred_username or not username in cred_username.lower():
                    continue

            last_written = datetime.fromtimestamp(os.path.getmtime(path))
            yield [cred_service, cred_username, last_written, path]

        
    def get_username(service):
        service = _sanitize_part(service)
        for cred in enumerate():
            if cred[0] == service:
                return cred[1]

        # not found in credential manager
        return None

    
    _not_in_the_password_store_re = re.compile(r"^Error: .+ is not in the password store.$")
    def get_password(service, username):
        key = _get_key(service, username)
        cp = subprocess.run(["pass", "show", key], capture_output=True, text=True)
        if cp.returncode != 0:
            if cp.returncode == 1 and _not_in_the_password_store_re.match(cp.stderr):
                return None
            else:
                raise SubprocessError(cp)
        return cp.stdout.splitlines()[0]


    def set_password(service, username, password):
        if not service:
            raise ValueError("service cannot be empty")

        password = password.splitlines()[0]
        inp = '%s\n' % password
        inp *= 2
        
        key = _get_key(service, username)
        cp = subprocess.run(['pass', 'insert', '--force', key], input=inp, capture_output=True, text=True)
        if cp.returncode != 0:
            raise SubprocessError(cp)


    def delete_password(service, username):
        key = _get_key(service, username)
        cp = subprocess.run(['pass', 'rm', '--force', key], capture_output=True)
        if cp.returncode != 0:
            raise SubprocessError(cp)


def add_arguments(parser):
    parser.add_argument("action", nargs="?", choices=["list", "ls", "get", "set", "delete", "del", "rm"], default="list")
    parser.add_argument("service", nargs="?")
    parser.add_argument("username", nargs="?")


def handle(action="list", service=None, username=None):
    from tabulate import tabulate

    if action == "get":
        # Get a credential
        if not service:
            service = input("Service: ")
        if not username:
            username = input(f"Username for service \"{service}\": ")
        print(get_password(service, username))
    elif action == "set":
        # Set a credential
        if not service:
            service = input("Service: ")
        if not username:
            username = input(f"Username for service \"{service}\": ")
        set_password(service, username, getpass(f"Password for service \"{service}\" (username: \"{username}\"): "))
    elif action in ["delete", "del", "rm"]:
        # Delete a credential
        if not service:
            service = input("Service: ")
        if not username:
            username = input(f"Username for service \"{service}\": ")
        delete_password(service, username)
    else:
        # Search/list available credentials
        if sys.platform == "win32":
            print("See: rundll32.exe keymgr.dll, KRShowKeyMgr")
        else:
            print("See: find ~/.password-store -type f -name '*.gpg'")
            
        rows = []
        for cred in enumerate(service=service, username=username):
            rows.append(cred)

        rows.sort(key=lambda row: row[0].lower())

        # Display LastWritten as simplified local time
        try:
            i = enumerate_headers.index("LastWritten")            
            for row in rows:
                localdt = row[i].astimezone()
                row[i] = localdt.strftime("%H:%M" if localdt.date() == datetime.now().date() else "%Y-%m-%d")

        except ValueError:
            pass

        print(tabulate(rows, enumerate_headers))



import re, logging
from pathlib import Path
from .. import get_connection

_pg_msg_re = re.compile(r"^(?P<pglevel>[A-Z]+)\:\s(?P<message>.+(?:\r?\n.*)*)$", re.MULTILINE)

class PgLogHandler:
    """
    Usage example:
    ```
    from django.apps import AppConfig
    from django.db.backends.signals import connection_created
    from zut.db.postgresql import PgLogHandler

    def connection_created_receiver(sender, connection, **kwargs):
        if connection.alias == "default":
            connection.connection.notices = PgLogHandler(connection.alias)

    class MainConfig(AppConfig):
        default_auto_field = "django.db.models.BigAutoField"
        name = "main"
        
        def ready(self):
            connection_created.connect(connection_created_receiver)
    ```
    """
    def __init__(self, alias: str):
        self.logger = logging.getLogger(f"pg:{alias}")
    
    def append(self, fullmsg: str):
        fullmsg = fullmsg.strip()
        m = _pg_msg_re.match(fullmsg)
        if not m:
            self.logger.warning(fullmsg)
            return

        message = m.group("message").strip()
        pglevel = m.group("pglevel")
        if pglevel == "EXCEPTION":
            level = logging.ERROR
        elif pglevel == "WARNING":
            level = logging.WARNING
        else:
            level = logging.INFO

        if level <= logging.INFO and message.endswith("\" does not exist, skipping"):
            return

        self.logger.log(level, message)


_pg_reserved_words = None

def get_pg_reserved_words(connection = None) -> list[str]:
    global _pg_reserved_words
    
    connection = get_connection(connection=connection)

    if _pg_reserved_words is None:
        with connection.cursor() as cursor:
            cursor.execute("select word from pg_get_keywords() where catcode = 'R'")
            _pg_reserved_words = [row[0] for row in cursor.fetchall()]
    return _pg_reserved_words


_pg_types = None

def get_pg_types(connection = None) -> dict[int, str]:
    global _pg_types

    connection = get_connection(connection=connection)

    if _pg_types is None:
        with connection.cursor() as cursor:
            cursor.execute("select oid, typname from pg_type")
            _pg_types = {row[0]: row[1] for row in cursor.fetchall()}
    return _pg_types


# TODO/FIXME: does not seem to work
# def upgrade_sequence_values(connection: connection, start_with=1001, prefix=None, suffix="_id_seq"):
#     name_like = (prefix if prefix else '') + "%" + (suffix if suffix else '')
#     sequence_names = []
#     with connection.cursor() as cursor:
#         cursor.execute(SQL("select sequencename from pg_sequences where sequencename like {} and start_value = 1 and coalesce(last_value, 0) < {}").format(Literal(name_like), Literal(start_with)))
#         for record in cursor:
#             sequence_names.append(record[0])

#     for sequence_name in sequence_names:
#         with connection.cursor() as cursor:
#             cursor.execute(SQL("alter sequence {} start with {} restart with {}").format(Identifier(sequence_name), Literal(start_with), Literal(start_with)))



from __future__ import annotations
import logging
from django.core.management import base
from django.db import connection
from zut.db import deploy_sql, ZUT_DB_BASE_DIR, get_backend, seed_from_enums

logger = logging.getLogger(__name__)

class Command(base.BaseCommand):
    def handle(self, action=None, **kwargs):
        seed_from_enums()
        
        backend = get_backend(connection=connection)
        backend_dir = ZUT_DB_BASE_DIR.joinpath(backend.name.lower())
        deploy_sql(backend_dir)



from __future__ import annotations
import subprocess
from tempfile import TemporaryDirectory
from pathlib import Path
from .format import SubprocessError
from .net import determine_proxy_url

def download_gpg_key(keyid: str, target_path: Path, keyserver: str = None):
    with TemporaryDirectory() as tmpdir:
        # Retrieve the key
        cmd = ["gpg", "--homedir", tmpdir]
        
        if keyserver:
            cmd += ["--keyserver", keyserver]
            
            if keyserver.startswith("hkp://"):
                proxy_url = determine_proxy_url(keyserver)
                if proxy_url:
                    cmd += ["--keyserver-options", f"http-proxy={proxy_url}"]

        cmd += ["--recv-keys", keyid]
        subprocess.run(cmd, capture_output=True, text=True, check=True)

        # Export the key
        cmd = ["gpg", "--homedir", tmpdir, "--output", target_path, "--export", keyid]
        subprocess.run(cmd, capture_output=True, text=True, check=True)

def verify_gpg_signature(sign_path: Path, public_key_path: Path):
    cmd = ["gpg", "--no-default-keyring", "--keyring", public_key_path, "--verify", sign_path]
    cp = subprocess.run(cmd, capture_output=True, text=True)
    if cp.returncode != 0:
        raise SubprocessError(cp)



from __future__ import annotations
import logging, os, socket, base64, re, urllib.request
from typing import Callable
from urllib.parse import unquote, urlparse
from contextlib import closing
from unittest import TestCase

logger = logging.getLogger(__name__)

def check_socket(host, port, timeout=1):
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as sock:
        sock.settimeout(timeout)
        try:
            returncode = sock.connect_ex((host, port))
            if returncode == 0:
                return True
            else:
                logger.debug("socket connnect_ex returned %d", returncode)
                return False
        except Exception as e:
            logger.debug("socket connnect_ex: %s", e)
            return False


class SimpleProxyHandler(urllib.request.BaseHandler):
    # Inspired from urllib.request.ProxyHandler   
    handler_order = 100 # Proxies must be in front

    def __init__(self, host: str, port: str, username: str = None, password: str | Callable[[], str] = None, scheme: str = "http"):
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.scheme = scheme


    def finalize(self) -> str:
        # Determine password if this is a callable
        if self.password and isinstance(self.password, Callable):
            self.password = self.password()
        
        # Build proxy url (complete, and safe for logging)
        proxy_url = self.scheme + "://"
        proxy_logurl = self.scheme + "://"

        if self.username:
            proxy_url += self.username
            proxy_logurl += self.username

            if self.password:
                proxy_url += ":" + self.password
                proxy_logurl += ":" + ("*" * len(self.password))

            proxy_url += "@"
            proxy_logurl += "@"

        proxy_url += f"{self.host}:{self.port}"
        proxy_logurl += f"{self.host}:{self.port}"
        logger.debug(f"use proxy for urllib: {proxy_logurl}")

        # Check proxy existency
        if not check_socket(self.host, self.port):
            logger.warning(f"cannot connect to proxy for urllib: {proxy_logurl}")

        # Prepare usefull variables for add_header and set_proxy
        self.hostport = unquote(f"{self.host}:{self.port}")

        if self.username:
            if not self.password:
                raise ValueError(f"missing password for urllib proxy: {proxy_logurl}")
            userpass = '%s:%s' % (unquote(self.username), unquote(self.password))
            self.authorization = "Basic " + base64.b64encode(userpass.encode()).decode("ascii")
        else:
            self.authorization = None

        return proxy_url


    def http_open(self, req):
        if not req.host:
            return None
        
        if urllib.request.proxy_bypass(req.host):
            # NOTE: because Proxy-Authorization header is not encrypted, we must add it ONLY when we're actually talking to the proxy
            return None

        if not hasattr(self, "authorization"):
            self.finalize()

        if self.authorization:
            req.add_header("Proxy-Authorization", self.authorization)
        req.set_proxy(self.hostport, self.scheme)
        return None
    

    def https_open(self, req):
        return self.http_open(req)


def detect_proxy_settings():
    # Detect proxy URL
    if "HTTP_PROXY" in os.environ:
        url = os.environ["HTTP_PROXY"]
    elif "http_proxy" in os.environ:
        url = os.environ["http_proxy"]
    else:
        return None

    # Detect proxy exclusions
    if "NO_PROXY" in os.environ:
        exclusions = os.environ["NO_PROXY"]
    elif "no_proxy" in os.environ:
        exclusions = os.environ["no_proxy"]
    else:
        exclusions = None

    # Parse proxy URL
    o = urlparse(url)
    m = re.match(r"^(?:(?P<username>[^\:]+)(?:\:(?P<password>[^\:]+))?@)?(?P<host>[^@\:]+)\:(?P<port>\d+)$", o.netloc)
    if not m:
        logger.error("cannot register proxy: invalid proxy netloc \"%s\"" % o.netloc)
        return None

    proxy_settings = {}
    proxy_settings["host"] = m.group("host")
    proxy_settings["port"] = int(m.group("port"))
    proxy_settings["username"] = m.group("username")
    proxy_settings["password"] = m.group("password")
    proxy_settings["scheme"] = o.scheme
    proxy_settings["exclusions"] = exclusions
    return proxy_settings            


_proxy_handler: SimpleProxyHandler = None
_proxy_hostport: str = None
_proxy_exclusions: list[str] = None

def configure_proxy(**proxy_settings):
    """
    Configure proxy for urllib and winhttp (on windows) requests.
    """
    global _proxy_handler, _proxy_hostport, _proxy_exclusions

    if not proxy_settings:
        proxy_settings = detect_proxy_settings()

    # Stop here if no proxy
    if not proxy_settings:
        return

    host = proxy_settings["host"]
    port = proxy_settings["port"]
    username = proxy_settings.get("username", None)
    password = proxy_settings.get("password", None)
    scheme = proxy_settings.get("scheme", "http")
    exclusions = proxy_settings.get("exclusions", None)

    _proxy_hostport = f"{host}:{port}"
    _proxy_exclusions = exclusions

    for key in ["HTTP_PROXY", "http_proxy", "HTTPS_PROXY", "https_proxy"]:
        if key in os.environ:
            del os.environ[key]

    # Ensure no_proxy is specified (must be passed as environment variable for urllib)
    if exclusions:
        if isinstance(exclusions, list):
            exclusions = ",".join(exclusions)
        os.environ["NO_PROXY"] = exclusions

    # Register proxy for urllib
    try:
        _proxy_handler = SimpleProxyHandler(host=host, port=port, username=username, password=password, scheme=scheme)
        opener = urllib.request.build_opener(_proxy_handler)
        urllib.request.install_opener(opener)
    except Exception as e:
        logger.error("cannot register proxy for urllib: %s", e)


def determine_proxy_url(url: str):
    """
    Determine proxy URL associated to the given target URL.
    """
    o = urlparse(url)
    if urllib.request.proxy_bypass(o.netloc if o.netloc else o.path):
        return None
    if not _proxy_handler:
        return None
    return _proxy_handler.finalize()


def test_urllib_request(case: TestCase, url: str, expected_regex: str, label=None, timeout: int = 2) -> str:
    if label is None:
        label = url
    msg=f"{label} urllib"

    try:
        res = urllib.request.urlopen(url, timeout=timeout)
        text = res.read().decode('utf-8')
        case.assertRegex(text, expected_regex, msg=msg)
    except urllib.error.URLError as e:
        case.fail(f"{msg}: {e.reason}")



from __future__ import annotations
import logging, re, sys, os
from pathlib import Path
from psycopg2.sql import SQL, Literal
from django.conf import settings
from django.db import connection
from django.core.management import base, call_command, get_commands
from zut.env import get_venv
from zut.db import get_backend, Backend, BackendNotSupported

logger = logging.getLogger(__name__)

_migration_name_re = re.compile(r"^(\d+)_")

class Command(base.BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument("--drop", dest="action", action="store_const", const="drop")
        parser.add_argument("--bak", dest="action", action="store_const", const="bak")
        parser.add_argument("--bak-to", dest="action")
        parser.add_argument("remake_migrations_for_apps", nargs="*", help="apps for which migrations are remade")

    def handle(self, action: str = None, remake_migrations_for_apps: list[str] = [], **kwargs):
        if not settings.DEBUG:
            raise ValueError("reinit may be used only in DEBUG mode")
        if not action:
            raise ValueError("please confirm what to do with current data: --drop, --bak or --bak-to")

        backend = get_backend(connection=connection)
        if backend != Backend.POSTGRESQL:
            raise BackendNotSupported(backend)

        self.REMAKE_MIGRATIONS_AFTER: dict[str,int] = getattr(settings, "REMAKE_MIGRATIONS_AFTER", {})
        self.BASE_DIR: Path = settings.BASE_DIR
        self.remake_migrations_for_apps = remake_migrations_for_apps

        if action == "drop":
            self.drop()
        else:
            self.move_to_schema(action)

        self.delete_nonmanual_migrations()
        self.rename_manual_migrations()

        logger.info("make migrations")
        call_command("makemigrations")

        self.restore_renamed_migrations()
        
        logger.info("migrate")
        call_command("migrate")

        logger.info("createsuperuser")
        call_command("createsuperuser", "--noinput")

        defined_commands = get_commands()

        if "seed" in defined_commands:
            logger.info("seed")
            call_command("seed")


    def move_to_schema(self, new_schema, old_schema="public"):
        sql = """do language plpgsql
    $$declare
        old_schema name = {};
        new_schema name = {};
        sql_query text;
    begin
        sql_query = format('create schema %I', new_schema);

        raise notice 'applying %', sql_query;
        execute sql_query;
    
        for sql_query in
            select
                format('alter %s %I.%I set schema %I', case when table_type = 'VIEW' then 'view' else 'table' end, table_schema, table_name, new_schema)
            from information_schema.tables
            where table_schema = old_schema
            and table_name not in ('geography_columns', 'geometry_columns', 'spatial_ref_sys') -- postgis
        loop
            raise notice 'applying %', sql_query;
            execute sql_query;
        end loop;
    end;$$;
    """

        with connection.cursor() as cursor:
            cursor.execute(SQL(sql).format(Literal(old_schema), Literal(new_schema if new_schema else "public")))


    def drop(self, schema="public"):
        sql = """do language plpgsql
    $$declare
        old_schema name = {};
        sql_query text;
    begin
        -- First, remove foreign-key constraints
        for sql_query in
            select
                format('alter table %I.%I drop constraint %I', table_schema, table_name, constraint_name)
            from information_schema.table_constraints
            where table_schema = old_schema and constraint_type = 'FOREIGN KEY'
            and table_name not in ('geography_columns', 'geometry_columns', 'spatial_ref_sys') -- postgis
        loop
            raise notice 'applying %', sql_query;
            execute sql_query;
        end loop;

        -- Then, drop tables
        for sql_query in
            select
                format('drop %s if exists %I.%I cascade'
                    ,case when table_type = 'VIEW' then 'view' else 'table' end
                    ,table_schema
                    ,table_name
                )
            from information_schema.tables
            where table_schema = old_schema
            and table_name not in ('geography_columns', 'geometry_columns', 'spatial_ref_sys') -- postgis
        loop
            raise notice 'applying %', sql_query;
            execute sql_query;
        end loop;
    end;$$;
    """

        with connection.cursor() as cursor:
            cursor.execute(SQL(sql).format(Literal(schema)))

    def should_remake_migration(self, path: Path, remake_manuals=False):
        if not remake_manuals and path.name.endswith("_manual.py"):
            return False

        if get_venv() in path.parents:
            # exclude migrations located in ".venv" directory
            return False

        app_name = path.parent.parent.name
        if self.remake_migrations_for_apps and not app_name in self.remake_migrations_for_apps:
            return False

        if self.REMAKE_MIGRATIONS_AFTER and app_name in self.REMAKE_MIGRATIONS_AFTER:
            m = _migration_name_re.match(path.name)
            if not m:
                return False

            migration_number = int(m.group(1))
            if migration_number <= self.REMAKE_MIGRATIONS_AFTER[app_name]:
                return False
        
            else:
                return True

        elif self.remake_migrations_for_apps and app_name in self.remake_migrations_for_apps:
            return True

        else:
            return False

    def delete_nonmanual_migrations(self):
        """ Delete non-manual migrations """
        for path in self.BASE_DIR.glob("*/migrations/0*.py"):
            if self.should_remake_migration(path):
                logger.info(f"delete {path}")
                path.unlink()
            else:
                logger.info(f"preserve {path}")

    def rename_manual_migrations(self):
        self.renamed: dict[Path,Path] = {}

        """ Rename manual migrations to py~ """
        for path in self.BASE_DIR.glob("*/migrations/*_manual.py"):
            if self.should_remake_migration(path, remake_manuals=True):
                target = path.with_name(f"{path.name}~")
                logger.info(f"rename {path} to {target}")
                path.rename(target)
                self.renamed[path] = target
            else:
                logger.info(f"preserve {path}")
    
    def restore_renamed_migrations(self):
        """ Restore migrations from py~ """
        for origin, renamed in self.renamed.items():
            logger.info(f"rename {renamed} to {origin}")
            renamed.rename(origin)



from __future__ import annotations
import logging, win32com.client, win32inetcon, pywintypes
from unittest import TestCase
from . import _proxy_hostport, _proxy_exclusions

logger = logging.getLogger(__name__)

_proxy_winhttp_exclusions: str = None

def create_winhttp_request(timeout=None):
    global _proxy_winhttp_exclusions

    winhttp_req = win32com.client.Dispatch('WinHTTP.WinHTTPRequest.5.1')

    if timeout:
        winhttp_req.SetTimeouts(str(timeout*1000), str(timeout*1000), str(timeout*1000), str(timeout*1000))

    if _proxy_hostport:
        # Ensure _proxy_winhttp_exclusions
        # See: https://docs.microsoft.com/en-us/windows/win32/winhttp/iwinhttprequest-setproxy
        # NOTE: no need to pass credentials, this is handled directly by Windows
        # NOTE: remove CIDR-block exclusions (not supported)
        if not _proxy_winhttp_exclusions:
            if _proxy_exclusions:
                _proxy_winhttp_exclusions = ",".join([exclusion for exclusion in _proxy_exclusions.split(",") if not "/" in exclusion])
            else:
                logger.warning("no exclusions for proxy: missing NO_PROXY environment variable?")
                _proxy_winhttp_exclusions = "localhost"

        HTTPREQUEST_PROXYSETTING_DEFAULT   = 0
        HTTPREQUEST_PROXYSETTING_PRECONFIG = 0
        HTTPREQUEST_PROXYSETTING_DIRECT    = 1
        HTTPREQUEST_PROXYSETTING_PROXY     = 2
        winhttp_req.SetProxy(HTTPREQUEST_PROXYSETTING_PROXY, _proxy_hostport, _proxy_winhttp_exclusions)
        winhttp_req.SetAutoLogonPolicy(HTTPREQUEST_PROXYSETTING_DEFAULT)

    return winhttp_req


def test_winhttp_request(case: TestCase, url: str, expected_regex: str, label=None, timeout: int = 2) -> str:
    if label is None:
        label = url
    msg=f"{label} winhttp"

    winhttp_req = create_winhttp_request(timeout=timeout)
    winhttp_req.Open('GET', url, False)
    try:
        winhttp_req.Send()
        winhttp_req.WaitForResponse()
        if winhttp_req.Status == 200:
            text = winhttp_req.ResponseText
            case.assertRegex(text, expected_regex, msg=msg)
        else:
            case.fail(f"{msg}: {winhttp_req.Status} {winhttp_req.StatusText}")
    except pywintypes.com_error as e:
        if e.excepinfo[5] + 2**32 & 0xffff == win32inetcon.ERROR_INTERNET_TIMEOUT:
            details = "timed out"
        else:
            details = e.excepinfo[2].strip()
        case.fail(f"{msg}: {details}")



from __future__ import annotations
from .cmd import add_commands, exec_command, call_command
from .env import get_venv
from .format import human_bytes, slugen, ExtendedJSONDecoder, ExtendedJSONEncoder, SubprocessError
from .git import get_git_tags, get_git_hash, check_git_version_tags
from .gpg import download_gpg_key, verify_gpg_signature
from .log import configure_logging
from .net import configure_proxy
from .out import open_out



from __future__ import annotations
from cmarkgfm import github_flavored_markdown_to_html
from django.views.generic import TemplateView
from django.template.loader import get_template
from html.parser import HTMLParser

class MarkdownTemplateView(TemplateView):
    extends = "base.html"

    def render_to_response(self, context, **response_kwargs):
        markdown_template = get_template(self.template_name)
        markdown_text = markdown_template.render(context, self.request)
        html_content = github_flavored_markdown_to_html(markdown_text)

        new_context = {}
        new_context["extends"] = self.extends
        new_context["title"] = MarkdownTemplateView.find_html_tag_content(html_content, "h1")
        new_context["html_content"] = html_content
        return super().render_to_response(new_context, **response_kwargs)

    def get_template_names(self):
        return ["_content.html"] # located in main/templates

    @staticmethod
    def find_html_tag_content(html: str, tag: str):        
        class HTMLTagParser(HTMLParser):
            def __init__(self, searched_tag):
                super().__init__()
                self.searched_tag = searched_tag
                self.match = False
                self.data = None

            def handle_starttag(self, tag, attributes):
                self.match = tag == self.searched_tag

            def handle_data(self, data):
                if self.match:
                    self.data = data
                    self.match = False

        parser = HTMLTagParser(tag)
        parser.feed(html)
        return parser.data



from __future__ import annotations
import logging
from enum import Enum
from pathlib import Path
from ..format import slugen

try: # usage with Django
    from django.db import router, connections
    from django.core.exceptions import ImproperlyConfigured
    from django.apps import apps
    with_django = True
except: # usage without Django
    with_django = False


logger = logging.getLogger(__name__)

ZUT_DB_BASE_DIR = Path(__file__).parent

class Backend(Enum):
    POSTGRESQL = 1


class BackendNotSupported(ValueError):
    def __init__(self, info: Backend|type):
        if isinstance(info, Backend):
            message = f"backend not supported: {info.name}"
        elif isinstance(info, type):
            message = f"connection type not supported: {info.__module__}.{info.__name__}"
        else:
            message = info
        super().__init__(message)


def get_backend(connection):
    search = type(connection).__module__ + "." + type(connection).__name__
    if search == "django.utils.connection.ConnectionProxy":
        search = connection._connections[connection._alias].vendor
    elif hasattr(connection, "vendor"):
        # e.g. django.db.backends.postgresql.base.DatabaseWrapper, django.contrib.gis.db.backends.postgis.base.DatabaseWrapper
        search = connection.vendor

    if search in ["postgresql", "psycopg2.extensions.connection"]:
        return Backend.POSTGRESQL
    else:
        raise BackendNotSupported(type(connection))


def get_connection(connection = None, model: type = None, for_write: bool = False):
    if connection:
        if model:
            raise ValueError("connection and model options cannot be both used")
        if not isinstance(connection, str):
            return connection

    if not with_django:
        raise ValueError("usage without connection option requires Django")

    try:
        if isinstance(connection, str):
            alias = connection

        elif model:
            if not hasattr(model, "objects"):
                raise ValueError(f"type {model.__name__} does not seem to be a Django model")

            alias = router.db_for_write(model) if for_write else router.db_for_read(model)
            if not alias:
                alias = "default"

        else:
            alias = "default"

        return connections[alias]

    except ImproperlyConfigured as err:
        raise ValueError(f"Django improperly configured: please provide \"connection\" option")


def dictfetchall(cursor) -> list[dict]:
    """
    Return all rows from a cursor as a dict.
    """ 
    desc = cursor.description 
    return [
            dict(zip([col[0] for col in desc], row)) 
            for row in cursor.fetchall() 
    ]


def call_procedure(*names: str, connection = None):
    connection = get_connection(connection=connection)
    backend = get_backend(connection)

    for name in names:
        if backend == Backend.POSTGRESQL:
            from psycopg2.sql import SQL, Identifier
            sql = SQL("call {}()").format(Identifier(name))
        else:
            raise BackendNotSupported(backend)

        with connection.cursor() as cursor:
            logger.info("call %s", name)
            cursor.execute(sql)


def truncate_table(*names: str|type, connection = None):
    for name in names:
        if isinstance(name, str):
            model = None
            db_table = name
        else:
            # 'name' is assumed to be a model class    
            model = name
            db_table = model._meta.db_table
        
        connection = get_connection(connection=connection, model=model, for_write=True)
        backend = get_backend(connection)
        if backend == Backend.POSTGRESQL:
            from psycopg2.sql import SQL, Identifier
            sql = SQL("truncate table {}").format(Identifier(db_table))
        else:
            raise BackendNotSupported(backend)
        
        with connection.cursor() as cursor:
            logger.info(f"truncate table {db_table}")
            cursor.execute(sql)


def seed_from_enum(model: type, enum: type|str = "Values", value_field: str="id", name_field: str="name", slug_field: str="slug", attr_fields: list[str]=None):
    if not hasattr(model, "objects"):
        raise ValueError(f"type {model.__name__} does not seem to be a Django model")
    if isinstance(enum, str):
        enum = getattr(model, enum, None)
        if not enum:
            raise ValueError("model %s has no %s attribute" % (model, enum))
    if not isinstance(enum, type):
        raise ValueError(f"invalid type for enum: {type(enum).__name__}")
    if not issubclass(enum, Enum):
        raise ValueError(f"not a subclass of Enum: {enum.__name__}")

    has_slug_field = False
    fields = model._meta.get_fields()
    for field in fields:
        if field.name == slug_field:
            has_slug_field = True
            break

    logger.info(f"seed {model.__module__}.{model.__name__}.{enum.__name__}")
    for literal in enum:
        name = literal.name.lower().replace("_", " ").capitalize()

        defaults = {
            name_field: name
        }

        if has_slug_field:
            defaults[slug_field] = slugen(name)

        if attr_fields:
            for attr, field in attr_fields.items():
                defaults[field] = getattr(literal, attr)

        kwargs = {
            value_field: literal.value,
            "defaults": defaults
        }
        
        model.objects.get_or_create(**kwargs)


def seed_from_enums(*app_names: str):
    if not with_django:
        raise ValueError("Django required")

    for app in apps.get_app_configs():
        if app_names and app.name not in app_names:
            continue

        for model in app.get_models():
            if hasattr(model, "Values"):
                seed_from_enum(model)


def deploy_sql(*paths: Path|str, encoding = "utf-8", connection = None):
    connection = get_connection(connection=connection)

    actual_paths: list[Path] = []
    for path in paths:
        if isinstance(path, str):
            path = Path(path)
        actual_paths.append(path)

    actual_paths.sort()

    for path in actual_paths:
        if path.is_dir():
            subpaths = sorted(path.iterdir())
            deploy_sql(*subpaths, encoding=encoding, connection=connection)

        elif not path.name.endswith(".sql"):
            continue # ignore

        elif path.name.endswith("_revert.sql"):
            continue # ignore

        else:
            logger.info("execute %s", path)
            sql = path.read_text(encoding=encoding)
            with connection.cursor() as cursor:
                cursor.execute(sql)


def revert_sql(*paths: Path|str, encoding = "utf-8", connection=None):
    connection = get_connection(connection=connection)

    actual_paths: list[Path] = []
    for path in paths:
        if isinstance(path, str):
            path = Path(path)
        actual_paths.append(path)

    actual_paths.sort(reverse=True)

    for path in actual_paths:
        if path.is_dir():
            subpaths = sorted(path.iterdir())
            revert_sql(*subpaths, encoding=encoding, connection=connection)

        elif not path.name.endswith("_revert.sql"):
            continue # ignore

        else:
            logger.info("execute %s", path)
            sql = path.read_text(encoding=encoding)
            with connection.cursor() as cursor:
                cursor.execute(sql)



import logging
from django.http import HttpResponseForbidden
from django.core.exceptions import ImproperlyConfigured

try:
    from django.contrib.auth.mixins import AccessMixin
    from django.contrib.auth.views import LoginView, LogoutView, redirect_to_login
    from django.views.generic.base import RedirectView
    improperly_configured_error = None
        
    try:
        from rest_framework.views import APIView
        with_rest_framework = True
    except ImportError:
        with_rest_framework = False
except ImproperlyConfigured as improperly_configured_error:
    # delay exception in class instantiation to avoid failure during unittest discovery
    pass


logger = logging.getLogger(__name__)

class StaffAuthorizationMiddleware:
    no_default_for = ["/", "/status"]

    def __init__(self, get_response):
        if improperly_configured_error:
            raise improperly_configured_error
        self.get_response = get_response

    def __call__(self, request):
        return self.get_response(request)

    def process_view(self, request, view_func, view_args, view_kwargs):
        if request.path in self.no_default_for:
            return # no default permission

        if view_func.__module__.startswith("django.contrib.admin."):
            return # no default permission

        if not hasattr(view_func, "view_class"):
            # function-based view
            if not self.is_authorized(request.user):
                logger.debug("function-based view %s: default authorization required", ".".join([view_func.__module__, view_func.__name__]))
                return self._deny_or_login(request)
            
        else:
            # class-based view
            view = view_func.view_class

            if issubclass(view, (LoginView, LogoutView, RedirectView)):
                return # no default permission

            if with_rest_framework and issubclass(view, APIView):
                # API view (Django Rest Framework)
                if not view.permission_classes:
                    if not self.is_authorized(request.user):
                        logger.debug("no permission_classes for rest_framework view %s: default authorization required required", ".".join([view.__module__, view.__name__]))
                        return HttpResponseForbidden() # do not redirect to login page (this is supposed to be accessed by javascript or as an API)

            else:
                # Standard class-based view
                if not issubclass(view, AccessMixin):
                    if not self.is_authorized(request.user):
                        logger.debug("no AccessMixin for view %s: default authorization required", ".".join([view.__module__, view.__name__]))
                        return self._deny_or_login(request)

    def is_authorized(self, user):
        return user.is_staff

    def _deny_or_login(self, request):
        if request.user.is_authenticated:
            return HttpResponseForbidden()
        else:
            return redirect_to_login(next=request.get_full_path())


class AdminAuthorizationMiddleware(StaffAuthorizationMiddleware):
    def is_authorized(self, user):
        return user.is_admin



from __future__ import annotations
import os, sys, csv, locale
from pathlib import Path
from io import IOBase

# Register CSV dialect for French version of Excel
class ExcelFr(csv.excel):
    delimiter = ";"

csv.register_dialect('excel-fr', ExcelFr())

class open_out:
    """
    ## Usage examples
    
    ### Export text either on stdout or on csv file

    ```
    with open_out(filename or "stdout") as o:
        o.file.write("10: Zidane")
    ```
    
    ### Export tabular data either on stdout or on csv file

    ```
    with open_out(filename or "stdout") as o:
        o.append_headers("Id", "Name")
        o.append_row(10, "Zidane")
    ```
    """
    file: IOBase

    def __init__(self, out: str|Path|IOBase = None, append: bool = False, newline: str = None, encoding: str = None, csv_dialect: str = None, **kwargs):
        self._append = append
        self._newline = newline
        self._encoding = encoding
        self._csv_dialect = csv_dialect

        # Update out
        self._file_opened = False
        self.file: IOBase = None
        self._strpath: str = None
        if out != "noop":
            if not out or out == "stdout":
                self.file = sys.stdout
            elif out == "stderr":
                self.file = sys.stderr
            elif isinstance(out, IOBase):
                self.file = out
            elif isinstance(out, Path):
                self._strpath = str(out)
            elif isinstance(out, str):
                self._strpath = out
            else:
                raise AttributeError(f"out: invalid type {type(out).__name__}")

        # For CSV file:
        # - Set newline to '', otherwise newlines embedded inside quoted fields will not be interpreted correctly. See footnote of: https://docs.python.org/3/library/csv.html
        # - Set encoding to utf-8-sig (UTF8 with BOM): CSV is for exchanges, encoding should not depend on the exporting operating system. BOM is necessary for correct display with Excel
        if (self._strpath and self._strpath.lower().endswith(".csv")) or self._csv_dialect:
            if self._newline is None:
                self._newline = ''
            if self._encoding is None:
                self._encoding = 'utf-8-sig'

        # Handle strpath
        if self._strpath:
            # Replace "{key}" in path by keyword arguments
            for key, value in kwargs.items():
                self._strpath = self._strpath.replace("{"+key+"}", value)
        elif self.file:
            if hasattr(self.file, "name"):
                self._strpath = self.file.name
            if not self._strpath:
                self._strpath = f"<{type(self.file).__name__}>"

    def __enter__(self):
        if not self.file and self._strpath:
            self._file_opened = True
            Path(self._strpath).parent.mkdir(parents=True, exist_ok=True)
            self.file = open(self._strpath, "a" if self._append else "w", newline=self._newline, encoding=self._encoding)
        return self

    def __exit__(self, *args):
        if not self.file and not self._strpath:
            return # noop
        
        if self._file_opened:
            # CSV to file
            if self.headers:
                self.csv_writer.writerow(self.headers)
            
            if self.rows:
                for row in self.rows:
                    while len(row) < len(self.headers):
                        row.append(None)
                    self.csv_writer.writerow(row)

            self.file.close()
        
        else:
            # Tabulate to IOBase
            if self.headers or self.rows:
                from tabulate import tabulate
                print(tabulate(self.rows, headers=self.headers), file=self.file)

    def __str__(self) -> str:
        if self._strpath:
            return self._strpath
        else:
            return "<noop>"

    # -------------------------------------------------------------------------
    # For tabular data
    # -------------------------------------------------------------------------

    @property
    def headers(self):
        if not hasattr(self, "_headers"):
            return None
        return self._headers
    
    def append_headers(self, *args):
        headers = args[0] if len(args) == 1 and isinstance(args[0], list) else args
        if hasattr(self, "_headers"):
            self._headers += headers
        else:
            self._headers = headers

    @property
    def rows(self) -> list[list]:
        if not hasattr(self, "_rows"):
            self._rows = []
        return self._rows

    def append_row(self, *args):
        if not self.file and not self._strpath:
            # noop
            return
        
        if len(args) == 1 and isinstance(args[0], dict):
            # input contains header and value
            if not hasattr(self, "_headers"):
                self._headers = []
            
            row = [None] * len(self._headers)
            
            for header, value in args[0].items():
                try:
                    index = self._headers.index(header)
                    row[index] = value
                except:
                    self._headers.append(header)
                    row.append(value)

        elif len(args) == 1 and isinstance(args[0], list):
            row = args[0]
        else:
            row = args

        self.rows.append(row)

    @property
    def csv_writer(self):
        if not hasattr(self, "_csv_writer"):
            dialect = os.environ.get("CSV_DIALECT", self._csv_dialect)
            if not dialect:
                loc = locale.getdefaultlocale()
                if loc[0].startswith("fr"):
                    dialect = "excel-fr"
                else:
                    dialect = "excel"
            self._csv_writer = csv.writer(self.file, dialect=dialect)
        return self._csv_writer



from __future__ import annotations
import re, json, unicodedata, subprocess
from datetime import datetime, date


# -----------------------------------------------------------------------------
# Colors
# -----------------------------------------------------------------------------

FOREGROUND_RED="\033[0;31m%s\033[0;0m"
FOREGROUND_GREEN="\033[0;32m%s\033[0;0m"
FOREGROUND_GRAY="\033[0;90m%s\033[0;0m"
BACKGROUND_RED="\033[0;41m%s\033[0;0m"


# -----------------------------------------------------------------------------
# Decimal numbers
# -----------------------------------------------------------------------------

class _PowerValue():
    def __init__(self, value, unit='B', base="decimal", maxrank=None):
        self.base = base
        if base == "binary":
            power = 1024
            self.rank_labels = []
            for rank in ['', 'Ki', 'Mi', 'Gi', 'Ti']:
                self.rank_labels.append(rank)
                if maxrank and rank != '' and maxrank.upper() == rank[0].upper():
                    break
        elif base == "decimal":
            power = 1000
            self.rank_labels = []
            for rank in ['', 'k', 'M', 'G', 'T']:
                self.rank_labels.append(rank)
                if maxrank and rank != '' and maxrank.upper() == rank[0].upper():
                    break
        else:
            raise ValueError(f'invalid base: {base}')

        self.unit = unit
        self.value = value
        self.rank = 0

        if self.value is None:
            return

        while self.value > power and self.rank < len(self.rank_labels) - 1:
            self.value /= power
            self.rank += 1
        
    def __str__(self):
        if self.value is None:
            return ''
        return ('{:,.1f}' if self.rank > 0 else '{:,.0f}').format(self.value) + ' ' + self.rank_labels[self.rank] + self.unit

def human_bytes(value, base="binary", maxrank=None):
    if value is None:
        return ''

    if base not in ["binary", "decimal", "both"]:
        raise ValueError("invalid base '%s'" % base)
        
    if base in ["binary", "both"]:
        binary_str = str(_PowerValue(value, base="binary", maxrank=maxrank))
        if base == "binary":
            return binary_str

    if base in ["decimal", "both"]:
        decimal_str = str(_PowerValue(value, base="decimal", maxrank=maxrank))
        if base == "decimal":
            return decimal_str
    
    if base == "both":
        return '%-12s' % binary_str + ' (' + decimal_str + ')'


# -----------------------------------------------------------------------------
# Strings
# -----------------------------------------------------------------------------

def slugen(value, separator='-') -> str:
    """ 
    Generate a slug.
    Simplier and more determinist than django.utils.text.slugify().
    """
    if value is None:
        return ""
    
    # Remove accents and other diacritic/non-ascii characters
    value = unicodedata.normalize("NFKD", str(value)).encode("ascii", "ignore").decode("ascii")

    # Lowercase the string
    value = value.lower()

    # Replace everything that is not a letter or digit by hyphens
    value = re.sub(r"[^a-z0-9]", "-", value)

    # Trim leading, trailing, and consecutive hyphens
    return re.sub(r"-+", separator, value).strip("-")


# -----------------------------------------------------------------------------
# JSON
# -----------------------------------------------------------------------------

def extended_json_decode_hook(obj):
    if isinstance(obj, dict):
        for key, value in obj.items():
            obj[key] = extended_json_decode_hook(value)

    elif isinstance(obj, list):
        for i, value in enumerate(obj):
            obj[i] = extended_json_decode_hook(value)

    elif isinstance(obj, str):
        if len(obj) < 10:
            return obj # ignore
        
        if re.match(r'^\d{4}-\d{2}-\d{2}$', obj):
            # date only
            return datetime.strptime(obj, "%Y-%m-%d").date()
        
        m = re.match(r'^(\d{4}-\d{2}-\d{2}T)(\d{2}:\d{2}:\d{2})(\.\d{3,6})?(Z|[\+\-]\d{2}:\d{2})?$', obj)
        if not m:
            return obj # ignore

        datepart = m.group(1) # mandatory
        timepart = m.group(2) # mandatory
        microsecondpart = m.group(3) # optional
        timezone = m.group(4) # optional
        
        # adapt timezone: replace 'Z' with +0000, or +XX:YY with +XXYY
        if timezone == 'Z':
            timezone = '+0000'
        elif timezone:
            timezone = timezone[:-3] + timezone[-2:]
        
        # NOTE: we don't decode XX:XX:XX into a time: too much risky that it's not actually a time
        # if not datepart:
        #     # time only: we only handle non-timezone-aware times, see: DjangoJSONEncoder
        #     if timezone:
        #         return obj

        #     if microsecondpart:
        #         return time.strptime(f"{timepart}{microsecondpart}", "%H:%M:%S.%f")
        #     else:
        #         return time.strptime(f"{timepart}", "%H:%M:%S")

        # datetime
        if microsecondpart:
            return datetime.strptime(f"{datepart}{timepart}{microsecondpart}{timezone}", "%Y-%m-%dT%H:%M:%S.%f%z")
        else:
            return datetime.strptime(f"{datepart}{timepart}{timezone}", "%Y-%m-%dT%H:%M:%S%z")

    return obj


class ExtendedJSONDecoder(json.JSONDecoder):
    """
    JSONDecoder subclass that knows how to decode date/time, decimal types, and UUIDs.
    Reverse of: ExtendedJSONEncoder.
    Usage example: json.loads(data, cls=ExtendedJSONDecoder)
    """
    def __init__(self, **kwargs):
        if not 'object_hook' in kwargs:
            kwargs['object_hook'] = extended_json_decode_hook
        super().__init__(**kwargs)


class ExtendedJSONEncoder(json.JSONEncoder):
    """
    Adapted from: django.core.serializers.json.DjangoJSONEncoder
    Usage example: json.dumps(data, indent=2, cls=ExtendedJSONEncoder)
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def default(self, o):
        # See "Date Time String Format" in the ECMA-262 specification.
        if isinstance(o, datetime):
            r = o.isoformat()
            if o.microsecond and o.microsecond % 1000 == 0:
                r = r[:23] + r[26:]
            if r.endswith("+00:00"):
                r = r[:-6] + "Z"
            return r
        elif isinstance(o, date):
            return o.isoformat()
        # elif isinstance(o, datetime.time):
        #     if is_aware(o):
        #         raise ValueError("JSON can't represent timezone-aware times.")
        #     r = o.isoformat()
        #     if o.microsecond:
        #         r = r[:12]
        #     return r
        # elif isinstance(o, datetime.timedelta):
        #     return duration_iso_string(o)
        # elif isinstance(o, (decimal.Decimal, uuid.UUID, Promise)):
        #     return str(o)
        else:
            return super().default(o)


# -----------------------------------------------------------------------------
# Subprocess
# -----------------------------------------------------------------------------
class SubprocessError(subprocess.CalledProcessError):
    """
    Similar to CalledProcessError but do not show entire cmd (might contain sensitive information)
    and show stdout and stderr.
    """
    def __init__(self, cp: subprocess.CompletedProcess, label=None):
        self.label = label
        super().__init__(cp.returncode, cp.args, cp.stdout, cp.stderr)
    
    def __str__(self):
        msg = f"{self.label if self.label else self.cmd[0]}"
        if self.returncode != 0:
            msg += f" - returncode: {FOREGROUND_GRAY}" % (self.returncode)
        if self.stderr:
            msg += f" - stderr: {FOREGROUND_GRAY}" % (self.stderr.strip() if isinstance(self.stderr, str) else self.stderr)
        if self.stdout:
            msg += f" - stdout: {FOREGROUND_GRAY}" % (self.stdout.strip() if isinstance(self.stdout, str) else self.stdout)
        return msg



from __future__ import annotations
import argparse
from .log import configure_logging
from .cmd import add_command, exec_command

configure_logging()

parser = argparse.ArgumentParser()
subparsers = parser.add_subparsers()

add_command(subparsers, "zut.cred")
add_command(subparsers, "zut.git", name="checkversion")

if __name__ == "__main__":
    exec_command(parser)



from __future__ import annotations
import sys
from pathlib import Path

_venv: Path|None = None
_venv_lookedup = False

def get_venv() -> Path|None:
    """ Returns venv Path if any, else None """
    global _venv, _venv_lookedup

    if not _venv_lookedup:
        _venv = None
        _venv_lookedup = True
        cwd = Path.cwd()
        potential_venv = Path(sys.executable).parent.parent
        if cwd in potential_venv.parents:
            _venv = potential_venv

    return _venv



#!/usr/bin/env python3
"""
See: python setup.py --help-commands

python setup.py build        # build everything needed to install in "build/"
python setup.py sdist        # create a source distribution (tarball, zip file, etc.) in "dist/"
python setup.py bdist_wheel  # create wheel in "dist/"
"""
from pathlib import Path
from setuptools import setup, find_packages

BASE_DIR = Path(__file__).resolve().parent

VERSION = "0.3.4"

def parse_pip_requirements(filename="requirements.txt"):
    path = BASE_DIR.joinpath(filename)
    with open(path, "r", encoding="utf-8") as file:
        requirements = []
        for line in file.read().splitlines():
            pos = line.find("#")
            if pos >= 0:
                line = line[0:pos]
            line = line.strip()
            if not line:
                continue
            if line.startswith("-r "):
                continue # should be added as an EXTRA requirement
            requirements.append(line)
        return requirements

if __name__ == "__main__":
    setup(
        name="zut",
        version=VERSION,
        author="Ipamo",
        author_email="dev@ipamo.net",
        description="Reusable Python, Django and PostgreSql utilities",
        long_description=BASE_DIR.joinpath("README.md").read_text().strip(),
        long_description_content_type="text/markdown",
        url="https://gitlab.com/ipamo/zut",
        project_urls={
            "Bug Tracker": "https://gitlab.com/ipamo/zut/issues",
        },
        classifiers=[
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
        ],
        packages=["zut"] + [f"zut.{pkg}" for pkg in find_packages(where="./zut")],
        package_data={
            "zut.db.postgresql": ["*.sql"]
        },
        python_requires=">=3.7.3", # Debian 10 (Buster)
        setup_requires=[
            "setuptools"
        ],
        install_requires=[],
        extras_require={ "extra": parse_pip_requirements() },
        data_files=[
            ('', ["requirements.txt"]),
        ],
    )



import logging
from django import template
from django.utils.safestring import mark_safe
from django.templatetags.static import static
from django.conf import settings

logger = logging.getLogger(__name__)

register = template.Library()

# avoid logging warnings for every request
_missing_version: list[str] = []
_missing_integrity: list[str] = []

def _get_lib_url(package, file, version=None):
    LOCAL_STATIC_LIB = getattr(settings, "LOCAL_STATIC_LIB", False)
    if LOCAL_STATIC_LIB:
        return f"{static(LOCAL_STATIC_LIB if isinstance(LOCAL_STATIC_LIB, str) else 'lib')}/{package}/{file}"
    else:
        if version:
            return f"https://cdn.jsdelivr.net/npm/{package}@{version}/{file}"
        else:
            return f"https://cdn.jsdelivr.net/npm/{package}/{file}"


@register.simple_tag
def style_lib(package, file, version=None, integrity=None):
    """
    Usage example in Django base template:

        {% load static %}
        {% load static_lib %}
        ...
        <head>
        ...
        {% style_lib  'bootstrap' 'dist/css/bootstrap.min.css' '5.2.0' 'sha256-7ZWbZUAi97rkirk4DcEp4GWDPkWpRMcNaEyXGsNXjLg=' %}
        ...
        </head>
    """
    url = _get_lib_url(package, file, version)

    if not version and not url in _missing_version:
        logger.warning(f"missing version for style_lib: {url}")
        _missing_version.append(url)
        
    html = f"<link rel=\"stylesheet\" href=\"{url}\""
    
    if integrity:
        html += f" integrity=\"{integrity}\" crossorigin=\"anonymous\""
    else:
        logger.warning(f"missing integrity hash for style_lib: {url}")
        _missing_integrity.append(url)

    html += f" />"
    return mark_safe(html)


@register.simple_tag
def script_lib(package, file, version=None, integrity=None, defer=False):
    """
    Usage example in Django base template:

        {% load static %}
        {% load static_lib %}
        ...
        <head>
        ...
        {% script_lib 'bootstrap' 'dist/js/bootstrap.bundle.min.js' '5.2.0' 'sha256-wMCQIK229gKxbUg3QWa544ypI4OoFlC2qQl8Q8xD8x8=' %}
        ...
        </head>
    """
    url = _get_lib_url(package, file, version)
    
    if not version and not url in _missing_version:
        logger.warning(f"missing version for script_lib: {url}")
        _missing_version.append(url)

    html = f"<script"
    if defer:
        html=" defer"
    html += f" src=\"{url}\""
    
    if integrity:
        html += f" integrity=\"{integrity}\" crossorigin=\"anonymous\""
    elif not url in _missing_integrity:
        logger.warning(f"missing integrity hash for script_lib: {url}")
        _missing_integrity.append(url)

    html += f"></script>"
    return mark_safe(html)



"""
Add and execute commands easily, based on argparse.
Usefull for non-Django applications.
For Django applications, use including command management instead.
"""
from __future__ import annotations
import sys, argparse
from typing import Callable
from pathlib import Path
from importlib import import_module
from importlib.util import find_spec
from .format import FOREGROUND_RED

def add_command(subparsers: argparse._SubParsersAction, module: str, name: str = None):
    if isinstance(module, str):
        module = import_module(module)
    if not name:
        name = module.__name__.split(".")[-1]

    handle = getattr(module, "handle")
    add_arguments = getattr(module, "add_arguments", None)
    help = getattr(module, "HELP", None)

    subparser = subparsers.add_parser(name, help=help, description=help) # help: in list of commands, description: in command help
    subparser.set_defaults(func=handle)
    if add_arguments:
        add_arguments(subparser)


def add_commands(subparsers: argparse._SubParsersAction, package: str):
    package_spec = find_spec(package)
    if not package_spec:
        raise KeyError(f"package not found: {package}")
    if not package_spec.origin:
        raise KeyError(f"not a package: {package} (did you forget __init__.py ?)")
    package_path = Path(package_spec.origin).parent
    
    for module_path in package_path.iterdir():
        if module_path.is_dir() or module_path.name.startswith("_") or not module_path.name.endswith(".py"):
            continue

        module = module_path.stem
        add_command(subparsers, f"{package}.{module}")


def exec_command(parser: argparse.ArgumentParser, default_func: Callable = None):
    args = vars(parser.parse_args())
    func = args.pop("func", None)
    if not func:
        if default_func:
            func = default_func
        else:
            print(FOREGROUND_RED % "missing command name")
            sys.exit(2)

    r = func(**args)
    if not isinstance(r, int):
        r = 0 if r is None else 1
    sys.exit(r)


def call_command(cmd: str, *args: any, parser: argparse.ArgumentParser = None):
    if parser is None:
        from __main__ import parser

    cmd_with_args = [cmd] + [str(arg) for arg in args]
    args = vars(parser.parse_args(cmd_with_args))

    func = args.pop("func", None)
    if not func:
        raise ValueError("missing command name")

    return func(**args)



from __future__ import annotations
import subprocess, re, logging
from importlib import import_module
from .format import SubprocessError, FOREGROUND_RED

logger = logging.getLogger(__name__)

def get_git_tags(points_at: str = None, pattern: str = None) -> list[str]:
    """
    Get list of defined tags.
    Use points_at="HEAD" for tags pointing on last commit.
    """
    cmd = ['git', 'tag', '--list']
    if points_at:
        cmd.append("--points-at")
        cmd.append(points_at)
    if pattern:
        cmd.append(pattern)

    cp = subprocess.run(cmd, check=True, text=True, capture_output=True)
    tags = cp.stdout.strip()
    if not tags:
        return []
    
    return tags.splitlines()


def get_git_hash(ref: str = None) -> None|str:
    """
    Get commit hash of a reference (branch, tag, etc).
    Use ref="HEAD" for last commit and ref=None for last commit only if there is no changes.
    """
    cmd = ['git', 'rev-list', '-n', '1', ref if ref else "HEAD"]

    cp = subprocess.run(cmd, text=True, capture_output=True)
    if cp.returncode == 128:
        return None
    elif cp.returncode != 0:
        raise SubprocessError(cp)

    hash = cp.stdout.strip()
    if not ref:
        cp = subprocess.run(['git', 'status', '--porcelain'], text=True, capture_output=True)
        if cp.returncode != 0:
            raise SubprocessError(cp)
        if cp.stdout:
            return f"changes[{hash}]"

    return hash


def check_git_version_tags(version: str) -> bool:
    if not version:
        raise AttributeError("version: empty")
    
    ok = True

    # Check version
    if not re.match(r"^\d+\.\d+\.\d+(?:\-[a-z0-9\-]+)?$", version):
        logger.error(f"version \"{version}\" does not match required regex")
        ok = False

    # Compare version with git tags
    tags = get_git_tags(pattern="v*")
    if not tags or not f"v{version}" in tags:
        logger.error(f"tag v{version} not found")
        ok = False
    else:
        # Ensure corresponding version tag matches current hash
        tag_hash = get_git_hash(f"v{version}")
        if tag_hash:
            current_hash = get_git_hash()
            if tag_hash != current_hash:
                logger.error(f"version tag (v{version}) hash: {FOREGROUND_RED % tag_hash}, current hash: {FOREGROUND_RED % current_hash}")
            ok = False

    return ok


def add_arguments(parser):
    parser.add_argument("--module", action="store", default="setup", help="the module containing VERSION attribute")


def handle(module: str):
    importedmodule = import_module(module)
    
    version = getattr(importedmodule, "VERSION")
    if not check_git_version_tags(version):
        return 1



from __future__ import annotations
import logging, logging.config, os, atexit
from .format import BACKGROUND_RED, FOREGROUND_RED

class CountHandler(logging.Handler):
    def __init__(self, level=logging.WARNING):
        self.counts: dict[int, int] = {}
        atexit.register(self.print_counts)
        super().__init__(level=level)

    def print_counts(self):
        msg = ""

        levelnos = sorted(self.counts.keys(), reverse=True)
        for levelno in levelnos:
            msg += (", " if msg else "") + "%s: %d" % (logging.getLevelName(levelno), self.counts[levelno])

        if msg:
            print("Logged " + msg)

    def emit(self, record: logging.LogRecord):
        if record.levelno >= self.level:
            if not record.levelno in self.counts:
                self.counts[record.levelno] = 1
            else:
                self.counts[record.levelno] += 1

def configure_logging(config: dict = None, nocount = False, level = logging.INFO):
    if not config:
        config = {
            "version": 1,
            "disable_existing_loggers": False,
            "formatters": {
                "standard": { 
                    "format": "%(levelname)s [%(name)s] %(message)s"
                },
            },
            "handlers": {
                "console": {
                    "level": "DEBUG",
                    "class": "logging.StreamHandler",
                    "formatter": "standard",
                }
            },
            "loggers": {
                "": { # root logger
                    "handlers": ["console"],
                    "level": os.environ["LOGLEVEL"].upper() if "LOGLEVEL" in os.environ else logging.getLevelName(level),
                },
            },
        }

        if not nocount:
            config["handlers"]["count"] = {
                "level": "WARNING",
                "class": CountHandler.__module__ + '.' + CountHandler.__qualname__,
            }
            for logger in config["loggers"]:
                config["loggers"][logger]["handlers"].append("count")

    # Color logging levels equal or greater than WARNING
    logging.addLevelName(logging.WARNING, FOREGROUND_RED % logging.getLevelName(logging.WARNING))
    logging.addLevelName(logging.ERROR, BACKGROUND_RED % logging.getLevelName(logging.ERROR))
    logging.addLevelName(logging.CRITICAL, BACKGROUND_RED % logging.getLevelName(logging.CRITICAL))

    # Apply configuration
    logging.config.dictConfig(config)
